from collections import namedtuple
from dataclasses import dataclass
from typing import Iterable, NamedTuple, Optional, Tuple
import scipy

import torch
from crtl.components.ann_vector_kb import ANNCandidateGenerator
from crtl.lkb.lkb import Concept, LexicalKnowledgeBase
from crtl.s3_file_utils.s3_file_utils import cached_path
import pathlib
from crtl.lkb.lkb import LexicalKnowledgeBase
import random
from torch.utils.data import Dataset
import numpy as np


@dataclass
class LinkingExample:
    term: str
    synonym_term: str
    random_example_terms: frozenset[str]
    hard_example_terms: frozenset[str]

    @property
    def negative_example_terms(self):
        return self.random_example_terms.union(self.hard_example_terms)


# tfidf everything before
InfRow = namedtuple(
    "InfRow", ("term_embedding", "synonym_embedding", "negative_embeddings")
)


class EmbeddedTermDataset(Dataset):
    def __init__(
        self,
        lkb: LexicalKnowledgeBase,
        ann_cg: ANNCandidateGenerator,
        device,
        no_of_negative_terms=10,
        max_no_of_hard_terms=5,
    ):
        self._lkb = lkb
        self._sorted_concepts = [
            concept
            for concept in sorted(self._lkb.cui_to_concept.values())
            if len(concept.terms) >= 2
        ]
        self._no_of_negative_terms = no_of_negative_terms
        self._max_no_of_hard_terms = max_no_of_hard_terms
        self._sorted_terms = sorted(
            set(term for term in self._lkb.term_to_cuis.keys() if term)
        )
        self._ann_cg = ann_cg
        self._term_to_idx = {
            term: idx for idx, term in enumerate(self._ann_cg._term_index)
        }
        print(type(self._ann_cg._vectors))
        self._vectors = self.get_tfidf_tensor(self._ann_cg._vectors).to(device)
        print("getting examples...")
        self._linking_examples = self.get_linking_examples()
        print("dataset initialized.")

    @staticmethod
    def get_tfidf_tensor(vectors: scipy.sparse._csr.csr_matrix) -> torch.tensor:
        print("getting tensor...")
        vectors = vectors.tocoo()
        indices = np.vstack((vectors.row, vectors.col))

        i = torch.LongTensor(indices)
        v = torch.FloatTensor(vectors.data)
        shape = vectors.shape
        print("got tensor")
        return torch.sparse_coo_tensor(i, v, shape).to_dense()
        # return torch.sparse.FloatTensor(i, v, torch.Size(shape))

    def concept_has_enough_terms(self, concept):
        return len([term for term in concept.terms if term in self._term_to_idx]) >= 2

    def transform_example(self, linking_example: LinkingExample) -> InfRow:
        assert len(linking_example.negative_example_terms) == self._no_of_negative_terms

        def get_vector(term):
            term_idx = self._term_to_idx[term]
            # print(self._vectors[term_idx].shape)
            return self._vectors[term_idx]

        example = InfRow(
            get_vector(linking_example.term),
            get_vector(linking_example.synonym_term),
            torch.vstack(
                [
                    get_vector(negative_term)
                    for negative_term in linking_example.negative_example_terms
                ]
            ),
        )
        return example

    def get_linking_examples(self) -> Iterable[LinkingExample]:
        concepts = self._sorted_concepts
        terms, synonym_terms = zip(
            *[
                self.get_synonym_terms(concept)
                for concept in concepts
                if self.concept_has_enough_terms(concept)
            ]
        )
        (
            random_negative_terms_batch,
            hard_negative_terms_batch,
        ) = self.get_negative_terms(terms)
        return [
            self.transform_example(
                LinkingExample(
                    term, synonym_term, random_negative_terms, hard_negative_terms
                )
            )
            for term, synonym_term, random_negative_terms, hard_negative_terms in zip(
                terms,
                synonym_terms,
                random_negative_terms_batch,
                hard_negative_terms_batch,
            )
        ]

    # def get_random_concepts(self)->Concept:
    #     return random.sample(self._sorted_concepts,k=self._batch_size)
    def get_negative_terms(self, terms):
        hard_negative_terms = self.get_hard_negative_terms_from_model(terms)
        random_negative_terms = self.get_random_terms(terms, hard_negative_terms)
        return random_negative_terms, hard_negative_terms

    def get_random_terms(self, terms_batch, hard_negative_terms_batch) -> Iterable[str]:
        sampled_random_terms_batch = []
        for canonical_term, hard_negative_terms in zip(
            terms_batch, hard_negative_terms_batch
        ):
            no_random_terms = self._no_of_negative_terms - len(hard_negative_terms)
            while True:  # do-while
                sampled_terms = set(
                    term
                    for term in random.sample(self._sorted_terms, k=no_random_terms)
                    if term not in self._get_all_synonyms(canonical_term)
                    and term not in hard_negative_terms
                    and term in self._term_to_idx
                )
                if len(sampled_terms) == no_random_terms:
                    break
            sampled_random_terms_batch.append(sampled_terms)
        return sampled_random_terms_batch

    def get_synonym_terms(self, concept) -> Optional[Tuple[str, str]]:
        return random.sample(
            [term for term in concept.terms if term in self._term_to_idx], k=2
        )

    def _get_all_synonyms(self, term):
        return [
            term
            for term_list in [
                self._lkb.cui_to_concept[cui].terms
                for cui in self._lkb.term_to_cuis[term]
            ]
            for term in term_list
        ]

    def get_hard_negative_terms_from_model(
        self, terms: Iterable[str]
    ) -> Iterable[Iterable[str]]:
        ann_cg_results = self._ann_cg._get_neighbor_terms(
            terms, k=self._max_no_of_hard_terms * 5
        )
        hard_terms_batch = []
        for term, result in zip(terms, ann_cg_results):
            hard_terms = set()
            for matched_term, _ in result:
                # matched_term = random.sample(sorted(self._lkb.cui_to_concept[mention_candidate.concept_id].terms),k=1)[0]
                if matched_term in self._lkb.term_to_cuis:
                    if (
                        len(
                            self._lkb.term_to_cuis[matched_term].intersection(
                                self._lkb.term_to_cuis[term]
                            )
                        )
                        == 0
                        and len(hard_terms) < self._max_no_of_hard_terms
                    ):
                        hard_terms.add(matched_term)
            hard_terms_batch.append(hard_terms)
            # print(len(hard_terms))
        return hard_terms_batch

    def __len__(self):
        return len(self._linking_examples)

    def __getitem__(self, idx):
        return self._linking_examples[idx]


if __name__ == "__main__":
    print("GETTING EXAMPLES...")
    lkb = LexicalKnowledgeBase.from_gzipped_jsonl(
        cached_path("s3://databrickstrenddata/crtl/lkb/trendnet_2.gz")
    )
    ann_cg = ANNCandidateGenerator.from_disk(
        pathlib.Path(
            "/Users/coltonflowers/CRTL/src/crtl/components/term_embedder/trendnet_2_ann_cg"
        )
    )
    egs = EmbeddedTermDataset(lkb, ann_cg)

    # for eg in egs:
    #     print(eg)
