import pathlib
from typing import Iterable
from crtl.components.ann_vector_kb import ANNCandidateGenerator
from crtl.lkb.lkb import LexicalKnowledgeBase
from crtl.s3_file_utils.s3_file_utils import cached_path
from crtl.components.term_embedder.dataset import EmbeddedTermDataset, InfRow
from crtl.lkb.lkb import LexicalKnowledgeBase
import torch
from torch import nn
from torch.utils.data import DataLoader
import torch
from scipy.spatial import distance
import numpy as np
from numpy import dot
from numpy.linalg import norm

# Create a vectorizer.
class TermEmbeddingNN(nn.Module):
    def __init__(self, input_feature_length: int):
        super().__init__()
        print(input_feature_length)
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(input_feature_length, 2048),
            nn.ReLU(),
            nn.Linear(2048, 1024),
            nn.ReLU(),
            nn.Linear(1024, 1024),
            nn.ReLU(),
            nn.Linear(1024, 512),
        )

    def forward(self, x):
        if self.training:
            return self.linear_relu_stack(
                torch.cat(
                    (
                        torch.unsqueeze(x.term_embedding, 1),
                        torch.unsqueeze(x.synonym_embedding, 1),
                        x.negative_embeddings,
                    ),
                    1,
                )
            )  # B x eg x f_out
        else:
            return self.linear_relu_stack(x)


class TermEmbedder:
    def __init__(self, lkb: LexicalKnowledgeBase, ann_cg, model, device):

        self._num_feat = ann_cg._vectors.shape[1]
        self._model = model
        model.to(device)
        self._device = device
        self._lkb = lkb
        self._terms = [
            term for term in list(self._lkb.term_to_cuis.keys()) if term != None
        ]
        self._ann_cg = ann_cg

    @classmethod
    def initalize(cls, lkb, ann_cg, device):
        num_feat = ann_cg._vectors.shape[1]
        model = TermEmbeddingNN(num_feat)
        return cls(lkb, ann_cg, model, device)

    def create_loss_fn(self, num_negative_examples, batch_size):
        embedding_loss_fn = torch.nn.CosineEmbeddingLoss(margin=0.5)
        cos = nn.CosineSimilarity(dim=1, eps=1e-6)
        cos_2 = nn.CosineSimilarity(dim=2, eps=1e-6)

        def loss_fn(predictions):
            labels = torch.tensor(
                [1] + [-1] * (num_negative_examples), device=self._device
            ).repeat(predictions.shape[0])
            sample_term_embedding = (
                torch.squeeze(predictions[:, 0, :])
                .unsqueeze(1)
                .expand(-1, num_negative_examples + 1, -1)
            )  # B x (eg-1) x f_out
            remaining_term_embeddings = torch.squeeze(
                predictions[:, 1:, :]
            )  # B x (eg-1) x f_out
            # print(remaining_term_embeddings.shape,sample_term_embedding.shape)
            flattended_sample_term_embedding = torch.flatten(
                sample_term_embedding, start_dim=0, end_dim=1
            )  # (B*(eg-1))x f_out
            flattended_remaining_term_embeddings = torch.flatten(
                remaining_term_embeddings, start_dim=0, end_dim=1
            )  # (B*(eg-1))x f_out
            # print(flattended_sample_term_embedding.shape,flattended_remaining_term_embeddings.shape,labels.shape)
            return embedding_loss_fn(
                flattended_sample_term_embedding,
                flattended_remaining_term_embeddings,
                labels,
            )

        def maximize_distance_diff_loss_fn(predictions):
            sample_term_embedding = torch.squeeze(
                predictions[:, 0, :]
            )  # B x (eg-1) x f_out
            synonym_term_embedding = torch.squeeze(predictions[:, 1, :])
            synonym_sim = cos(sample_term_embedding, synonym_term_embedding)
            negative_term_embeddings = torch.squeeze(predictions[:, 2:, :])
            expanded_term_embeddings = sample_term_embedding.unsqueeze(1).expand(
                -1, num_negative_examples, -1
            )
            negative_sims = cos_2(negative_term_embeddings, expanded_term_embeddings)
            # print("negative_sims\n",negative_sims.shape,"expanded_term_embeddings\n",expanded_term_embeddings.shape,negative_term_embeddings.shape,synonym_sim.shape)
            loss = torch.mean(negative_sims) - torch.mean(synonym_sim)
            return loss

        return loss_fn

    @staticmethod
    def get_labels(num_negative_examples, batch_size):
        return torch.tensor([1] + [-1] * (num_negative_examples)).repeat(batch_size)

    def train(
        self,
        batch_size,
        no_of_negative_terms,
        max_no_of_hard_negatives,
        learning_rate,
        epochs,
    ):
        loss_fn = self.create_loss_fn(no_of_negative_terms, batch_size)
        optimizer = torch.optim.Adam(self._model.parameters(), lr=learning_rate)
        for t in range(epochs):

            print(f"epoch:{t}")
            self._model.train()
            dataset = EmbeddedTermDataset(
                self._lkb,
                self._ann_cg,
                self._device,
                no_of_negative_terms,
                max_no_of_hard_negatives,
            )
            dataloader = DataLoader(dataset, batch_size=batch_size)
            size = len(dataloader.dataset)
            for batch, tfidf_embeddings in enumerate(dataloader):
                # print(batch)
                tfidf_embeddings: InfRow
                # Compute prediction and loss
                pred = self._model(tfidf_embeddings)
                loss = loss_fn(pred)
                # print(loss)

                # Backpropagation
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

                if batch % 100 == 0:
                    loss, current = loss.item(), (batch + 1) * len(tfidf_embeddings[0])
                    print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")
            del dataloader
            del dataset
        self._term_embeddings = self.embed_terms(self._terms)

    def embed_terms(self, terms):
        tfidf_tensor = EmbeddedTermDataset.get_tfidf_tensor(
            self._ann_cg._vectorizer.transform(terms)
        )
        batch_size = 64
        self._model.eval()
        with torch.no_grad():
            # for i in range(0, len(terms), batch_size):
            #     tensor = tfidf_tensor[i:i+batch_size]
            #     print(tensor.shape)
            #     self._model(tensor.to(self._device)
            tensors = tuple(
                self._model(tfidf_tensor[i : i + batch_size].to(self._device))
                for i in range(0, len(terms), batch_size)
            )
            embedding_tensor = torch.cat(tensors).cpu().numpy()
        print(embedding_tensor.shape)
        return embedding_tensor

    def find_most_similar(self, term):
        target_vector = self.embed_terms([term])
        distances = distance.cdist(
            target_vector, self._term_embeddings, "cosine"
        ).squeeze()
        min_index = np.argmin(distances)
        print(distances.shape)
        min_distance = distances[min_index]
        max_similarity = 1 - min_distance
        return self._terms[min_index], max_similarity

    def give_similarity(self, term_one, term_two):
        vectors = term_embedder.embed_terms([term_one, term_two])
        target_vector = vectors[0, :].squeeze()
        query_vector = vectors[1, :].squeeze()
        return dot(target_vector, query_vector) / (
            norm(target_vector) * norm(query_vector)
        )


if __name__ == "__main__":

    device = torch.device("mps")
    learning_rate = 1e-3
    batch_size = 64
    epochs = 10
    no_of_negative_terms = 20
    max_no_of_hard_negatives = 10
    lkb = LexicalKnowledgeBase.from_gzipped_jsonl(
        cached_path("s3://databrickstrenddata/crtl/lkb/trendnet_2.gz")
    )
    ann_cg = ANNCandidateGenerator.from_disk(
        pathlib.Path(
            "/Users/coltonflowers/CRTL/src/crtl/components/term_embedder/trendnet_2_ann_cg"
        )
    )
    term_embedder = TermEmbedder.initalize(lkb, ann_cg, device)
    term_embedder.train(
        batch_size,
        no_of_negative_terms,
        max_no_of_hard_negatives,
        learning_rate,
        epochs,
    )
    print(term_embedder.find_most_similar("childhood obesity"))
    print(term_embedder.find_most_similar("cries easily"))
    print(term_embedder.find_most_similar("daily cough"))
    print(term_embedder.find_most_similar("heart"))
